// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file NonLinSolver.h */

#ifndef AMDIS_NONLINSOLVER_H
#define AMDIS_NONLINSOLVER_H

#include <string>

#include "Global.h"
#include "CreatorInterface.h"
#include "MatrixVector.h"
#include "DOFMatrix.h"
#include "OEMSolver.h"

namespace AMDiS {

  /** 
   * \ingroup Solver
   *
   * \brief
   * Pure virtual base class for Newton, NewtonS and Banach which all
   * solves non linear equation systems. Sub classes must implement the methods
   * \ref init, \ref exit and \ref nlsolve.
   */
  class NonLinSolver
  {
  public:
    /** \brief
     * constructor.
     * \param name name of this solver
     */
    NonLinSolver(const std::string &name_, 
		 OEMSolver *solver,
		 NonLinUpdater *updater)
      : name(name_),
	linSolver(solver),
	nonLinUpdater(updater),
	tolerance(1.e-4),
	maxIter(50),
	info(8),
	initial_residual(0.0),
	residual(0.0),
	usedNorm(NO_NORM)
    {
      Parameters::get(name + "->tolerance", tolerance);
      Parameters::get(name + "->max iteration", maxIter);
      Parameters::get(name + "->info", info);
      Parameters::get(name + "->norm", usedNorm);  
    }

    /** \brief
     * destructor
     */
    virtual ~NonLinSolver() {}

    /** \brief
     * solves the non linear system. Uses sub class methods
     */
    inline int solve(SolverMatrix<Matrix<DOFMatrix*> >& mat,
		     SystemVector &x, SystemVector &rhs) 
    {
      init();
      int result = nlsolve(mat, x, rhs);
      exit();
      return result;
    }

    inline void setTolerance(double tol)
    {
      tolerance = tol;
    }

    inline double getTolerance()
    {
      return tolerance;
    }

    inline OEMSolver* getLinearSolver()
    { 
      return linSolver; 
    }

    inline void setNonLinUpdater(NonLinUpdater *up) 
    {
      nonLinUpdater = up;
    }

    inline NonLinUpdater *getNonLinUpdater()
    {
      return nonLinUpdater;
    }

  protected:
    /** \brief
     * Allocates needed memory. Must be overriden in sub classes.
     */  
    virtual void init() = 0;

    /** \brief
     * Solves the non linear system. Must be overriden in sub classes.
     */  
    virtual int nlsolve(SolverMatrix<Matrix<DOFMatrix*> >& matVec,
			SystemVector& x, SystemVector& rhs) = 0;

    /** \brief
     * Frees needed memory. Must be overriden in sub classes.
     */  
    virtual void exit() = 0;
  
    virtual int solveLinearSystem(SolverMatrix<Matrix<DOFMatrix*> >& mat,
				  SystemVector &fh, SystemVector &x)
    {
      FUNCNAME("NonLinSolver::solveLinearSystem()");
      TEST_EXIT(linSolver)("no solver\n");
      return linSolver->solveSystem(mat, x, fh);
    }

  protected:
    std::string        name;             /**< \brief name of the solver */
    OEMSolver *linSolver;        /**< \brief linear solver*/
    NonLinUpdater *nonLinUpdater;    /**< \brief non linear updater */
    double             tolerance;        /**< \brief solver tolerance */
    int                maxIter;          /**< \brief maximal # of iterations */
    int                info;             /**< \brief info level */
    double             initial_residual; /**< \brief initial residual */
    double             residual;         /**< \brief current residual */
    Norm               usedNorm;         /**< \brief used norm */
  };

  // ============================================================================
  // ===== class NonLinSolverCreator ============================================
  // ============================================================================

  /** \brief
   * Interface for creators of concrete NonLinSolvers. 
   */
  class NonLinSolverCreator : public CreatorInterface<NonLinSolver>
  { 
  public:
    virtual ~NonLinSolverCreator() {}

    void setName(std::string name_) 
    { 
      name = name_; 
    }

    void setLinearSolver(OEMSolver *solver) 
    { 
      linearSolver = solver; 
    }

    void setNonLinUpdater(NonLinUpdater *updater) 
    {
      nonLinUpdater = updater;
    }

  protected:
    std::string name;
    OEMSolver *linearSolver;
    NonLinUpdater *nonLinUpdater;
  };

}

#include "Newton.h"

#endif // AMDIS_NONLINSOLVER_H

